/*
 * Distributed under the terms of the GNU General Public License v2
 * $Header: /var/cvsroot/gentoo-projects/hardened/policycoreutils-extra/src/runscript_selinux.c,v 1.16 2006/03/10 21:28:58 pebenito Exp $
 */

/*
 * This is a modified NSA run_init for use in Gentoo's /sbin/runscript.
 */

/************************************************************************
 * BUILD OPTIONS:
 *
 * option USE_PAM:
 *
 * Set the USE_PAM constant if you want to authenticate users via PAM.
 * If USE_PAM is not set, users will be authenticated via direct
 * access to the shadow password file.
 *
 * If you decide to use PAM must be told how to handle run_init.  A
 * good rule-of-thumb might be to tell PAM to handle run_init in the
 * same way it handles su, except that you should remove the pam_rootok.so
 * entry so that even root must re-authenticate to run the init scripts
 * in the proper context.
 *
 * If you choose not to use PAM, make sure you have a shadow passwd file
 * in /etc/shadow.  You can use a symlink if your shadow passwd file
 * lives in another directory.  Example:
 *   su
 *   cd /etc
 *   ln -s /etc/auth/shadow shadow
 *
 * If you decide not to use PAM, you will also have to make run_init
 * setuid root, so that it can read the shadow passwd file.
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>               /* for malloc(), realloc(), free() */
#include <pwd.h>                  /* for getpwuid() */
#include <sys/types.h>            /* to make getuid() and getpwuid() happy */
#include <sys/wait.h>		  /* for wait() */
#include <sys/stat.h>		  /* for struct stat and friends */
#include <getopt.h>               /* for getopt_long() form of getopt() */
#include <selinux/selinux.h>
#include <selinux/get_default_type.h>
#include <selinux/context.h>              /* for context-mangling functions */
#include <fcntl.h>
#include <ctype.h>
#include <limits.h>
#ifdef USE_AUDIT
#include <libaudit.h>
#endif

#define RUN_INIT_FILE "run_init_type"
#define INITRC_FILE "initrc_context"
#ifdef USE_PAM

/************************************************************************
 *
 * All PAM code goes in this section.
 *
 ************************************************************************/

#include <unistd.h>               /* for getuid(), exit(), getopt() */

#include <security/pam_appl.h>    /* for PAM functions */
#include <security/pam_misc.h>    /* for misc_conv PAM utility function */

#define SERVICE_NAME "run_init"   /* the name of this program for PAM */
				  /* The file containing the context to run 
				   * the scripts under.                     */

int authenticate_via_pam( const struct passwd * );

/* authenticate_via_pam()
 *
 * in:     p_passwd_line - struct containing data from our user's line in 
 *                         the passwd file.
 * out:    nothing
 * return: value   condition
 *         -----   ---------
 *           1     PAM thinks that the user authenticated themselves properly
 *           0     otherwise
 *
 * This function uses PAM to authenticate the user running this
 * program.  This is the only function in this program that makes PAM
 * calls.
 *
 */

int authenticate_via_pam( const struct passwd *p_passwd_line ) {

  int result = 0;    /* our result, set to 0 (not authenticated) by default */
  pam_handle_t *pam_handle;      /* opaque handle used by all PAM functions */

  /* This is a jump table of functions for PAM to use when it wants to *
   * communicate with the user.  We'll be using misc_conv(), which is  *
   * provided for us via pam_misc.h.                                   */
  struct pam_conv pam_conversation = {
    misc_conv,
    NULL
  };

  /* Make `p_pam_handle' a valid PAM handle so we can use it when *
   * calling PAM functions.                                       */
  if( PAM_SUCCESS != pam_start( SERVICE_NAME,
				p_passwd_line->pw_name,
				&pam_conversation,
				&pam_handle ) ) {
    fprintf( stderr, "failed to initialize PAM\n" );
    exit( -1 );
  }

  /* Ask PAM to authenticate the user running this program */
  if( PAM_SUCCESS == pam_authenticate(pam_handle,0) ) {
    result = 1;  /* user authenticated OK! */
  }

  /* We're done with PAM.  Free `pam_handle'. */
  pam_end( pam_handle, PAM_SUCCESS );
 
  return( result );

} /* authenticate_via_pam() */

#else /* else !USE_PAM */


/************************************************************************
 *
 * All shadow passwd code goes in this section.
 *
 ************************************************************************/

#include <unistd.h>                         /* for getuid(), exit(), crypt() */
#include <crypt.h>
#include <shadow.h>                         /* for shadow passwd functions */
#include <string.h>                         /* for strlen(), memset() */

#define PASSWORD_PROMPT "Password:"         /* prompt for getpass() */

int authenticate_via_shadow_passwd( const struct passwd * );

/* authenticate_via_shadow_passwd()
 *
 * in:     p_passwd_line - struct containing data from our user's line in 
 *                         the passwd file.
 * out:    nothing
 * return: value   condition
 *         -----   ---------
 *           1     user authenticated themselves properly according to the
 *                 shadow passwd file.
 *           0     otherwise
 *
 * This function uses the shadow passwd file to authenticate the user running
 * this program.
 *
 */

int authenticate_via_shadow_passwd( const struct passwd *p_passwd_line ) {

  struct spwd *p_shadow_line; /* struct derived from shadow passwd file line */
  char *unencrypted_password_s;        /* unencrypted password input by user */
  char *encrypted_password_s; /* user's password input after being crypt()ed */

  /* Make `p_shadow_line' point to the data from the current user's *
   * line in the shadow passwd file.                                */
  setspent();            /* Begin access to the shadow passwd file. */
  p_shadow_line = getspnam( p_passwd_line->pw_name );
  endspent();            /* End access to the shadow passwd file. */
  if( !( p_shadow_line ) ) {
    fprintf( stderr, "Cannot find your entry in the shadow passwd file.\n" );
    exit( -1 );
  }

  /* Ask user to input unencrypted password */
  if( ! ( unencrypted_password_s = getpass( PASSWORD_PROMPT ) ) ) {
    fprintf( stderr, "getpass cannot open /dev/tty\n" );
    exit( -1 );
  }

  /* Use crypt() to encrypt user's input password.  Clear the *
   * unencrypted password as soon as we're done, so it is not * 
   * visible to memory snoopers.                              */
  encrypted_password_s = crypt( unencrypted_password_s,
				p_shadow_line->sp_pwdp );
  memset( unencrypted_password_s, 0, strlen( unencrypted_password_s ) );

  /* Return 1 (authenticated) iff the encrypted version of the user's *
   * input password matches the encrypted password stored in the      *
   * shadow password file.                                            */
  return( !strcmp( encrypted_password_s, p_shadow_line->sp_pwdp ) );

} /* authenticate_via_shadow_passwd() */

#endif /* if/else USE_PAM */

/*
 * authenticate_user()
 *
 * Authenticate the user.
 *
 * in:		nothing
 * out:		nothing
 * return:	0 When success
 *		-1 When failure
 */
int authenticate_user() {

#define INITLEN 255
  struct passwd *p_passwd_line;      /* struct derived from passwd file line */
  uid_t uid;

  /*
   * Determine the Linux user identity to re-authenticate.
   * If supported and set, use the login uid, as this should be more stable.
   * Otherwise, use the real uid.
   * The SELinux user identity is no longer used, as Linux users are now
   * mapped to SELinux users via seusers and the SELinux user identity space
   * is separate.
   */
#ifdef USE_AUDIT
  uid = audit_getloginuid();
  if (uid == (uid_t)-1)
    uid = getuid();
#else
  uid = getuid();
#endif      

  p_passwd_line = getpwuid(uid);
  if (!p_passwd_line) {
    fprintf(stderr, "cannot find your entry in the passwd file.\n");
    return(-1);
  }
  
  printf("Authenticating %s.\n",p_passwd_line->pw_name);

  /* 
   * Re-authenticate the user running this program.
   * This is just to help confirm user intent (vs. invocation by
   * malicious software), not to authorize the operation (which is covered
   * by policy).  Trusted path mechanism would be preferred.
   */
#ifdef USE_PAM
  if( !authenticate_via_pam(p_passwd_line) ) {
#else /* !USE_PAM */
  if( !authenticate_via_shadow_passwd(p_passwd_line) ) {
#endif /* if/else USE_PAM */
    fprintf(stderr, "runscript: incorrect password for %s\n", p_passwd_line->pw_name);
    return(-1);
  }

  /* If we reach here, then we have authenticated the user. */
#ifdef CANTSPELLGDB
  printf( "You are authenticated!\n" );
#endif  

  return 0;

} /* authenticate_user() */

/*
 * get_init_context()
 *
 * Get the CONTEXT associated with the context for the init scripts.             *
 *
 * in:		nothing
 * out:		The CONTEXT associated with the context.
 * return:	0 on success, -1 on failure.
 */
int get_context_from_file(const char *filename, security_context_t *context) {

  FILE* fp;
  char  buf[255], *bufp;
  int buf_len;
  char context_file[PATH_MAX];

  memset(context_file, '\0', sizeof(context_file));
  snprintf(context_file, sizeof(context_file)-1, "%s/%s", selinux_contexts_path(), filename);
  fp = fopen(context_file, "r");
  if (!fp) {
    fprintf(stderr, "runscript_selinux.c: Could not open file %s\n", context_file);
    return -1;
  }

  while (1) {           /* loop until we find a non-empty line */

    if (!fgets(buf, sizeof buf, fp))
      break;

    buf_len = strlen(buf);
    if (buf[buf_len-1] == '\n')
      buf[buf_len-1] = 0;

    bufp = buf;
    while (*bufp && isspace(*bufp))
      bufp++;

    if( *bufp ) {
      *context = strdup(bufp);
      if (!(*context))
	goto out;
      fclose(fp);
      return 0;
    }
  }
out:
  fclose(fp);
  fprintf(stderr, "runscript_selinux.c: No context in file %s\n", context_file);
  return -1;

} /* get_context_from_file() */

/*****************************************************************************
 * run_init()                                                                *
 *****************************************************************************/
void run_init(int argc, char **argv) {

  security_context_t new_context;    /* context for the init script context  */
  security_context_t curr_context;   /* context we are currently in */

  /* use this to extract the type */
  context_t curr_con;
  const char *curr_type, *run_init_type;

  /*
   * Return, if selinux is disabled.
   */
  if( is_selinux_enabled() < 1 ) {
#ifdef CANTSPELLGDB
    printf("SELinux is disabled.  Skipping selinux_runscript().\n");
#endif
    return;
  }
  
  /*
   * Get the run_init type (knowing that security_context_t is char*)
   */
  if( get_context_from_file(RUN_INIT_FILE,(security_context_t*)&run_init_type) == 0 ) {
#ifdef CANTSPELLGDB
    printf("run_init type is %s\n", run_init_type);
#endif
  } else {
    /* assume a reasonable default, rather than bailing out */
    run_init_type="run_init_t";
    fprintf(stderr, "runscript_selinux.c: Assuming run_init type is %s\n", run_init_type);
  }

 /*
  * Get our current context.
  */
  if ( getcon(&curr_context) == 0 ) {
#ifdef CANTSPELLGDB
    printf("current context is %s\n", curr_context);
#endif
  } else {
    exit(1);
  }

  /* extract the type from the context */
  curr_con=context_new(curr_context);
  curr_type=context_type_get(curr_con);

 /*
  * Authenticate if we're in the run_init domain.  If we are not
  * in the run_init domain, then either we are already in initrc_t
  * or things are mislabeled, and this won't work right anyway.
  */
  if( strncmp(run_init_type,curr_type,strlen(run_init_type)) == 0 ) {
    if( authenticate_user() != 0 ) {
      fprintf(stderr, "Authentication failed.\n");
      exit(1);
    }

  /*
   * Get the context for the script to be run in.
   */
    if( get_context_from_file(INITRC_FILE,&new_context) == 0 ) { 
#ifdef CANTSPELLGDB
      printf("initrc context is %s\n", new_context);
#endif  
    } else {
      /* assume a reasonable default, rather than bailing out */
      new_context="system_u:system_r:initrc_t";
      fprintf(stderr, "runscript_selinux.c: Assuming initrc context is %s\n", new_context);
    }

   /*
    * Set the correct context since we're not in the right domain.
    */

    if (setexeccon(new_context) < 0) {
       fprintf(stderr, "Could not set exec context to %s.\n", new_context);
       exit(1);
    }

    freecon(new_context);

    /*
     * Do not execvp the command directly from run_init; since it would run
     * under with a pty under sysadm_devpts_t. Instead, we call open_init_tty,
     * which transitions us into initrc_t, which then spawns a new
     * process, that gets a pty with context initrc_devpts_t. Just
     * execvp or using a exec(1) recycles pty's, and does not open a new
     * one. 
     */
    /* NULL check can be removed when legacy support is removed. */
    if( argv != NULL ) {
      if ( execvp("/usr/sbin/open_init_pty", argv) ) {
        perror("execvp");
        exit(-1);
      }
    }
  }

  context_free(curr_con);
  freecon(curr_context);

} /* run_init() */

void selinux_runscript() {
	/* legacy support */
#ifdef CANTSPELLGDB
	printf("Using legacy run_init code.\n");
#endif
	fprintf(stderr,"Warning: Your runscript uses legacy SELinux support.\n");
	fprintf(stderr,"Please update the sys-apps/baselayout package!\n");
	run_init(0,NULL);
}

void selinux_runscript2(int argc, char **argv) {
#ifdef CANTSPELLGDB
	printf("Using new run_init code.\n");
#endif
	/* run_init */
	run_init(argc,argv);
	/* should only get this far if selinux
	 * is disabled or there is no setexec
	 */
}
